import argparse

def str_to_bool(s):
    return s.lower() in ('yes', 'true', 't', '1')


def read_command_line_arguments():
    """
    Reads the command line arguments.

    Partial parsing is used. This is because the main function is executed twice when using a cluster manager, with the
    second run having extra arguments. If partial parsing is not used, arguments from the configuration file that are
    not defined by the argparse will be stored as command line arguments in the manager submission script during the
    first pass and then fed to the argparse on the second pass. Hence, ignoring the unknown arguments using partial
    parsing is necessary for the current setup.

    Partial parsing:
        https://docs.python.org/3/library/argparse.html#partial-parsing

    Returns:
        Object containing the model's configuration.
    """

    parser = argparse.ArgumentParser(description='Command line arguments')

    # Required arguments
    required = parser.add_argument_group('Required named arguments')
    required.add_argument('--task', '-t', type=str, help='The name of the task', required=True)
    required.add_argument('--config', '-c', type=str, help='Configuration name in task/TASK_NAME/config', required=True)

    # Directory paths
    directories = parser.add_argument_group('Directory paths')
    directories.add_argument('--exp-dir', '--exp_dir', type=str, help='Experiment outputs save directory')
    directories.add_argument('--work-dir', '--work_dir', type=str, help='Working directory')
    directories.add_argument('--dataset-dir', '--dataset_dir', type=str, help='The dataset directory')
    directories.add_argument('--ckpt-zoo-dir', '--ckpt_zoo_dir', type=str, help='The checkpoint zoo directory')

    # Model module and definition
    model = parser.add_argument_group('Model module name and definition')
    model.add_argument('--definition', type=str, help='Class definition of the model')
    model.add_argument('--module', type=str, help='Name of the module in task/TASK_NAME/model')

    # Training arguments
    training_arguments = parser.add_argument_group('Training arguments')
    training_arguments.add_argument('--train', type=str_to_bool, help='Perform training')
    training_arguments.add_argument('--trial', type=int, help='The trial number')
    training_arguments.add_argument(
        '--resumable', type=str_to_bool, help='Resumable training. Automatic resubmission to cluster manager',
    )
    training_arguments.add_argument('--resume-epoch', '--resume_epoch', type=int, help='Epoch to resume training from')
    training_arguments.add_argument(
        '--resume-ckpt-path', '--resume_ckpt_path', type=str, help='Checkpoint to resume training from',
    )
    training_arguments.add_argument(
        '--warm-start-ckpt-path', '--warm_start_ckpt_path', type=str, help='Checkpoint for warm-starting',
    )
    training_arguments.add_argument('--monitor', type=str, help='Metric to monitor')
    training_arguments.add_argument(
        '--monitor-mode',
        '--monitor_mode',
        type=str,
        help='whether the monitored metric is to be maximised or minimised (''max'' or ''min'')',
    )

    # Test arguments
    test = parser.add_argument_group('Testing arguments')
    test.add_argument('--test', type=str_to_bool, help='Evaluate the model on the test set')
    test.add_argument('--test-epoch', '--test_epoch', type=int, help='Test epoch')
    test.add_argument('--test-ckpt-path', '--test_ckpt_path', type=str, help='Path to checkpoint to be tested')

    # PyTorch Lightning Trainer arguments
    trainer = parser.add_argument_group('PyTorch Lightning Trainer arguments')
    trainer.add_argument('--debug', type=str_to_bool, help='One mini-batch for training, validation, and testing')

    # Distributed computing arguments
    distributed = parser.add_argument_group('Distributed computing arguments')
    distributed.add_argument('--num-workers', '--num_workers', type=int, help='No. of workers per DataLoader & GPU')
    distributed.add_argument('--num-gpus', '--num_gpus', type=int, help='Number of GPUs per node')
    distributed.add_argument('--num-nodes', '--num_nodes', type=int, help='Number of nodes')

    # Cluster manager arguments
    cluster = parser.add_argument_group('Cluster manager arguments')
    cluster.add_argument('--memory', type=str, help='Amount of memory per node')
    cluster.add_argument('--time-limit', '--time_limit', type=str, help='Job time limit')
    cluster.add_argument('--submit', type=str_to_bool, help='Submit job to the cluster manager')
    cluster.add_argument('--qos', type=str, help='Quality of service')
    cluster.add_argument('--begin', type=str, help='When to begin the Slurm job, e.g. now+1hour')
    cluster.add_argument('--slurm-cmd-path', '--slurm_cmd_path', type=str)
    cluster.add_argument('--email', type=str, help='Email for cluster manager notifications')

    # System arguments
    system = parser.add_argument_group('System arguments')
    system.add_argument('--cuda-visible-devices', '--cuda_visible_devices', type=str, help='Visible CUDA devices')
    system.add_argument('--venv-path', '--venv_path', type=str, help='Path to ''bin/activate'' of the virtualenv')

    args, unknown = parser.parse_known_args()

    return args



# Copyright The PyTorch Lightning team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
CSV logger
----------

CSV logger for basic experiment logging that does not require opening ports

"""
import csv
import logging
import os
from argparse import Namespace
from typing import Any, Dict, Optional, Union

import torch

from pytorch_lightning.core.saving import save_hparams_to_yaml
from pytorch_lightning.loggers.base import LightningLoggerBase, rank_zero_experiment
from pytorch_lightning.utilities.logger import _add_prefix, _convert_params
from pytorch_lightning.utilities.rank_zero import rank_zero_only, rank_zero_warn

log = logging.getLogger(__name__)


class ExperimentWriter:
    r"""
    Experiment writer for CSVLogger.

    Currently supports to log hyperparameters and metrics in YAML and CSV
    format, respectively.

    Args:
        log_dir: Directory for the experiment logs
    """

    NAME_HPARAMS_FILE = "hparams.yaml"
    NAME_METRICS_FILE = "metrics.csv"

    def __init__(self, log_dir: str) -> None:
        self.hparams = {}
        self.metrics = []

        self.log_dir = log_dir
        if os.path.exists(self.log_dir) and os.listdir(self.log_dir):
            rank_zero_warn(
                f"Experiment logs directory {self.log_dir} exists and is not empty."
                " Previous log files in this directory will be deleted when the new ones are saved!"
            )
        os.makedirs(self.log_dir, exist_ok=True)

        self.metrics_file_path = os.path.join(self.log_dir, self.NAME_METRICS_FILE)
        if os.path.isfile(self.metrics_file_path):
            with open(self.metrics_file_path) as f:
                self.metrics = [{k: v for k, v in row.items()} for row in csv.DictReader(f, skipinitialspace=True)]

    def log_hparams(self, params: Dict[str, Any]) -> None:
        """Record hparams."""
        self.hparams.update(params)


    def log_metrics(self, metrics_dict: Dict[str, float], step: Optional[int] = None) -> None:
        """Record metrics."""

        def _handle_value(value):
            if isinstance(value, torch.Tensor):
                return value.item()
            return value

        if step is None:
            step = len(self.metrics)

        metrics = {k: _handle_value(v) for k, v in metrics_dict.items()}
        metrics["step"] = step
        self.metrics.append(metrics)


    def save(self) -> None:
        """Save recorded hparams and metrics into files."""
        hparams_file = os.path.join(self.log_dir, self.NAME_HPARAMS_FILE)
        save_hparams_to_yaml(hparams_file, self.hparams)

        if not self.metrics:
            return

        last_m = {}
        for m in self.metrics:
            last_m.update(m)
        metrics_keys = list(last_m.keys())

        with open(self.metrics_file_path, "w", newline="") as f:
            writer = csv.DictWriter(f, fieldnames=metrics_keys)
            writer.writeheader()
            writer.writerows(self.metrics)



class CSVLogger(LightningLoggerBase):
    r"""
    Log to local file system in yaml and CSV format.

    Logs are saved to ``os.path.join(save_dir, name, version)``.

    Example:
        >>> from pytorch_lightning import Trainer
        >>> from pytorch_lightning.loggers import CSVLogger
        >>> logger = CSVLogger("logs", name="my_exp_name")
        >>> trainer = Trainer(logger=logger)

    Args:
        save_dir: Save directory
        name: Experiment name. Defaults to ``'default'``.
        version: Experiment version. If version is not specified the logger inspects the save
            directory for existing versions, then automatically assigns the next available version.
        prefix: A string to put at the beginning of metric keys.
        flush_logs_every_n_steps: How often to flush logs to disk (defaults to every 100 steps).
    """

    LOGGER_JOIN_CHAR = "-"

    def __init__(
        self,
        save_dir: str,
        name: Optional[str] = "lightning_logs",
        version: Optional[Union[int, str]] = None,
        prefix: str = "",
        flush_logs_every_n_steps: int = 100,
    ):
        super().__init__()
        self._save_dir = save_dir
        self._name = name or ""
        self._version = version
        self._prefix = prefix
        self._experiment = None
        self._flush_logs_every_n_steps = flush_logs_every_n_steps

    @property
    def root_dir(self) -> str:
        """Parent directory for all checkpoint subdirectories.

        If the experiment name parameter is an empty string, no experiment subdirectory is used and the checkpoint will
        be saved in "save_dir/version"
        """
        return os.path.join(self.save_dir, self.name)

    @property
    def log_dir(self) -> str:
        """The log directory for this run.

        By default, it is named ``'version_${self.version}'`` but it can be overridden by passing a string value for the
        constructor's version parameter instead of ``None`` or an int.
        """
        # create a pseudo standard path
        version = self.version if isinstance(self.version, str) else f"version_{self.version}"
        log_dir = os.path.join(self.root_dir, version)
        return log_dir

    @property
    def save_dir(self) -> Optional[str]:
        """The current directory where logs are saved.

        Returns:
            The path to current directory where logs are saved.
        """
        return self._save_dir

    @property
    @rank_zero_experiment
    def experiment(self) -> ExperimentWriter:
        r"""

        Actual ExperimentWriter object. To use ExperimentWriter features in your
        :class:`~pytorch_lightning.core.lightning.LightningModule` do the following.

        Example::

            self.logger.experiment.some_experiment_writer_function()

        """
        if self._experiment:
            return self._experiment

        os.makedirs(self.root_dir, exist_ok=True)
        self._experiment = ExperimentWriter(log_dir=self.log_dir)
        return self._experiment

    @rank_zero_only
    def log_hyperparams(self, params: Union[Dict[str, Any], Namespace]) -> None:
        params = _convert_params(params)
        self.experiment.log_hparams(params)


    @rank_zero_only
    def log_metrics(self, metrics: Dict[str, float], step: Optional[int] = None) -> None:
        metrics = _add_prefix(metrics, self._prefix, self.LOGGER_JOIN_CHAR)
        self.experiment.log_metrics(metrics, step)
        if step is not None and (step + 1) % self._flush_logs_every_n_steps == 0:
            self.save()


    @rank_zero_only
    def save(self) -> None:
        super().save()
        self.experiment.save()


    @rank_zero_only
    def finalize(self, status: str) -> None:
        self.save()


    @property
    def name(self) -> str:
        """Gets the name of the experiment.

        Returns:
            The name of the experiment.
        """
        return self._name

    @property
    def version(self) -> int:
        """Gets the version of the experiment.

        Returns:
            The version of the experiment if it is specified, else the next version.
        """
        if self._version is None:
            self._version = self._get_next_version()
        return self._version

    def _get_next_version(self):
        root_dir = self.root_dir

        if not os.path.isdir(root_dir):
            log.warning("Missing logger folder: %s", root_dir)
            return 0

        existing_versions = []
        for d in os.listdir(root_dir):
            if os.path.isdir(os.path.join(root_dir, d)) and d.startswith("version_"):
                existing_versions.append(int(d.split("_")[1]))

        if len(existing_versions) == 0:
            return 0

        return max(existing_versions) + 1


from pytorch_lightning import Trainer
from pytorch_lightning.callbacks import (
    EarlyStopping,
    LearningRateMonitor,
    ModelCheckpoint,
)
from pytorch_lightning.loggers import NeptuneLogger
from pytorch_lightning.loggers.tensorboard import TensorBoardLogger
from pytorch_lightning.plugins.environments import SLURMEnvironment
from dlhpcstarter.tools.mods.logger import CSVLogger
from dlhpcstarter.utils import resume_from_ckpt_path
from typing import Optional
import inspect
import logging
logging.getLogger(
    "neptune.new.internal.operation_processors.async_operation_processor",
).setLevel(logging.CRITICAL)


def trainer_instance(
    monitor: str,
    monitor_mode: str,
    task: str,
    config: str,
    trial: int,
    early_stopping: bool = False,
    patience: int = 0,
    min_delta: float = 0.0,
    divergence_threshold: Optional[float] = None,
    exp_dir_trial: Optional[str] = None,
    resumable: bool = True,
    resume_epoch: Optional[int] = None,
    resume_ckpt_path: Optional[str] = None,
    sched_inter: Optional[str] = None,  # 'step', 'epoch', or None.
    save_top_k: int = 1,
    every_n_epochs: Optional[int] = 1,
    every_n_train_steps: Optional[int] = None,
    debug: bool = False,
    submit: bool = False,
    neptune_api_key: Optional[str] = None,
    neptune_username: Optional[str] = None,
    num_nodes: int = 1,
    num_gpus: Optional[int] = None,
    mbatch_size: Optional[int] = None,
    accumulated_mbatch_size: Optional[int] = None,
    deterministic: bool = True,
    num_sanity_val_steps: int = 0,
    loggers: Optional[list] = None,
    callbacks: Optional[list] = None,  # [RichProgressBar()]
    plugins: Optional[list] = None,
    **kwargs,
) -> Trainer:
    """
    Creates an instance of pytorch_lightning.Trainer using key callbacks and loggers. Also changes some
    defaults for the init of pytorch_lightning.Trainer.

    Parameters for pytorch_lightning.Trainer are described here:
        https://pytorch-lightning.readthedocs.io/en/stable/common/trainer.html#trainer-class-api
    These will be captured by kwargs and passed to pytorch_lightning.Trainer

    Argument/s:
        monitor - metric to monitor for EarlyStopping and ModelCheckpoint.
        monitor_mode - whether the metric to be monitored is to be maximised or
            minimised.
        task - name of the task.
        config - name of the configuration.
        trial - trial identifier.
        early_stopping - stop training when a monitored metric has stopped
            improving.
        patience - no. of epochs with no improvement after which training will
            be stopped.
        min_delta - minimum change in the monitored quantity to qualify as an
            improvement.
        divergence_threshold - stop training as soon as the monitored quantity becomes worse than this threshold.
        exp_dir_trial - experiment directory for the trial. All outputs are saved to this path.
        resumable - whether the last completed epoch is saved to enable resumable training.
        resume_epoch - the epoch to resume training from.
        resume_ckpt_path - resume training from the specified checkpoint.
        sched_inter - learning rate scheduler interval ('step' or 'epoch').
        save_top_k - best k models saved according to the monitored metric. If
            0, no models are saved. If -1, all models are saved.
        every_n_epochs - save model every n epochs.
        every_n_epochs - save model every n training steps.
        debug - training, validation, and testing are completed using one mini-batch.
        submit - submit to cluster manager.
        neptune_api_key - API key, found on neptune.ai, for NeptuneLogger.
        neptune_username - Username for on neptune.ai, for NeptuneLogger.
        num_nodes - number of nodes for the job.
        num_gpus - number of GPUs per node.
        mbatch_size - mini-batch size of dataloaders.
        accumulated_mbatch_size - desired accumulated mini-batch size.
        deterministic - ensures that the training is deterministic.
        num_sanity_val_steps - runs n validation batches before starting the training routine.
        loggers - loggers for Trainer.
        callbacks - callbacks for Trainer.
        plugins - plugins for Trainer.
        kwargs - keyword arguments for Trainer.
    """
    accumulate_grad_batches = None
    loggers = [] if loggers is None else loggers
    """
    Potential default callbacks to add:
        - SLURMEnvironment(auto_requeue=True)  # This could be used in place of the auto requeue in transmodal.cluster
    """
    callbacks = [] if callbacks is None else callbacks
    plugins = [] if plugins is None else plugins
    if submit:
        plugins.append(SLURMEnvironment(auto_requeue=False))

    # Loggers
    loggers.append(CSVLogger(exp_dir_trial, name='', version=''))
    loggers.append(TensorBoardLogger(exp_dir_trial, name='', version='', default_hp_metric=False))
    if neptune_api_key is not None:
        custom_run_id = f'{config[:16]}_trial_{trial}'
        assert len(custom_run_id) <= 32, '"custom_run_id" must be less than or equal to 32 characters'
        assert neptune_username is not None, 'You must specify your neptune.ai username.'
        loggers.append(
            NeptuneLogger(
                api_key=neptune_api_key,
                project=f'{neptune_username}/{task.replace("_", "-")}',
                prefix='log',
                custom_run_id=custom_run_id,
                capture_stdout=False,
                capture_stderr=False,
                capture_hardware_metrics=True,
                capture_traceback=False,
                log_model_checkpoints=False,
                source_files=[],
                flush_period=30,
            )
        )

    # Model checkpointing
    assert (every_n_epochs is not None) or (every_n_train_steps is not None), 'Neither "every_n_epochs" or ' \
        '"every_n_train_steps" is set. No checkpoints will be saved.'
    assert save_top_k != 0, '"save_top_k" is 0, therefore, no checkpoints will be saved.'

    if every_n_epochs:
        callbacks.append(
            ModelCheckpoint(
                dirpath=exp_dir_trial,
                monitor=monitor,
                mode=monitor_mode,
                save_top_k=save_top_k,
                every_n_epochs=every_n_epochs,
                filename='{epoch:d}-{' + monitor + ':f}',
                save_last=True,
            )
        )

    if every_n_train_steps:
        if resumable:
            raise ValueError('Cannot resume training from a checkpoint that ended before the epoch ended. Fault '
                             'tolerant training needs to be implemented for this: '
                             'https://pytorch-lightning.readthedocs.io/en/latest/clouds'
                             '/fault_tolerant_training_expert.html#enable-fault-tolerant-behavior-on-your-own-cluster')
        callbacks.append(
            ModelCheckpoint(
                dirpath=exp_dir_trial,
                monitor=monitor,
                mode=monitor_mode,
                save_top_k=save_top_k,
                every_n_train_steps=every_n_train_steps,
                filename='{step:d}-{' + monitor + ':f}',
                save_last=False,  # cannot resume from this checkpoint.
            )
        )

    # Early stopping
    if early_stopping:
        callbacks.append(
            EarlyStopping(
                monitor=monitor,
                mode=monitor_mode,
                min_delta=min_delta,
                patience=patience,
                divergence_threshold=divergence_threshold,
                verbose=False,
            )
        )

    # Learning rate monitor
    if sched_inter is not None:
        callbacks.append(LearningRateMonitor(logging_interval=sched_inter))

    # Accumulate gradient batches
    if accumulated_mbatch_size:
        devices = num_gpus * num_nodes if num_gpus * num_nodes > 0 else 1
        accumulate_grad_batches = accumulated_mbatch_size / (mbatch_size * devices)
        assert accumulate_grad_batches.is_integer(), f'Desired accumulated_mbatch_size ({accumulated_mbatch_size}) ' \
                                                     f'can not be attained with mbatch_size={mbatch_size}, num_gpus=' \
                                                     f'{num_gpus}, and num_nodes={num_nodes}'
        accumulate_grad_batches = int(accumulate_grad_batches)

    # Resume from checkpoint
    ckpt_path = resume_from_ckpt_path(exp_dir_trial, resumable, resume_epoch, resume_ckpt_path)

    # Remove keyword arguments not associated with pytorch_lightning.Trainer.
    # Parameters associated with pytorch_lightning.Trainer:
    #   https://pytorch-lightning.readthedocs.io/en/stable/common/trainer.html#trainer-class-api
    kwargs = {k: v for k, v in kwargs.items() if k in inspect.signature(Trainer).parameters}

    # pytorch_lightning.Trainer
    return Trainer(
        logger=loggers,
        callbacks=callbacks,
        plugins=plugins,
        fast_dev_run=debug,
        gpus=num_gpus,
        num_nodes=num_nodes,
        accumulate_grad_batches=accumulate_grad_batches,
        deterministic=deterministic,
        num_sanity_val_steps=num_sanity_val_steps,
        resume_from_checkpoint=ckpt_path,
        **kwargs,
    )



from argparse import Namespace
from hydra import compose, initialize_config_dir
from pathlib import Path
from typing import Any, Dict, Optional, Tuple
import glob
import GPUtil
import importlib
import numpy as np
import os
import re
import warnings


def importer(
        definition: str,
        module: str,
        work_dir: Optional[str] = None,
):
    """
    Loads an instance of the definition from the given module. Can also load relative to a specified working directory.

    Argument/s:
        definition - the definition in the module. Typically the name of a function or class.
        module - the module, e.g. tasks.mnist.datamodule
        work_dir - working directory; loads the module relative to this directory.

    Returns:
        An instance of 'definition'.
    """

    if work_dir:

        absolute_path = os.path.join(work_dir, *module.split('.')) + '.py'

        assert os.path.isfile(absolute_path), \
            f'''{absolute_path} does not exist. The target definition was: {definition}.'''

        module = importlib.machinery.SourceFileLoader(module, absolute_path).load_module()

    else:

        path = os.path.join(*module.split('.')) + '.py'

        assert os.path.isfile(path), f'{path} does not exist. The target definition was: {definition}.'

        module = importlib.import_module(module)

    return getattr(module, definition)


def load_config_and_update_args(args: Namespace, print_args: bool = False) -> None:
    """
    Loads the configuration .yaml file and updates the args object.

    Argument/s:
        args - object that houses the arguments for the job.
        print_args - print the arguments for the job.
    """
    # Add the working directory to paths
    if not args.work_dir:
        args.work_dir = os.getcwd()

    # Configuration name
    if args.config.endswith('.yaml'):
        args.config_file_name = args.config
        args.config = args.config.replace('.yaml', '')
    else:
        args.config_file_name = args.config + '.yaml'

    # Load configuration using Hydra's Compose API
    args.config_dir = os.path.join(args.work_dir, 'task', args.task, 'config')
    with initialize_config_dir(
            version_base=None, config_dir=args.config_dir, job_name=f'{args.config}'
    ):
        config = compose(config_name=args.config)

    # Update args with config and check for conflicts
    args.config_full_path = os.path.join(args.config_dir, args.config_file_name)
    for k, v in config.items():
        if getattr(args, k, None) is None:
            setattr(args, k, v)
        else:
            if getattr(args, k) != v:
                raise ValueError(f'There is a conflict between command line argument "--{k} {getattr(args, k)}" '
                                    f'({type(getattr(args, k))}) and configuration argument "{k}: {v}" from '
                                    f'{args.config_full_path} ({type(v)}).')

    # The model name must be defined in the configuration or in the command line arguments
    assert args.module, f'"module" must be specified as a command line argument or in {args.config_full_path}.'
    assert args.definition, f'"definition" must be specified as a command line argument or in {args.config_full_path}.'
    assert args.exp_dir, f'"exp_dir" must be specified as a command line argument or in {args.config_full_path}.'

    # Defaults: There is probably a better place to do this
    args.num_workers = args.num_workers if args.num_workers is not None else 1
    args.num_nodes = args.num_nodes if args.num_nodes is not None else 1
    if args.submit:
        args.time_limit = args.time_limit if args.time_limit is not None else '02:00:00'
        args.memory = args.memory if args.memory is not None else '16GB'

    # Add the task, configuration name, and the trial number to the experiment directory
    args.trial = args.trial if args.trial is not None else 0
    args.exp_dir_trial = os.path.join(args.exp_dir, args.task, args.config, 'trial_' + f'{args.trial}')
    Path(args.exp_dir_trial).mkdir(parents=True, exist_ok=True)

    if print_args:
        print(f'args: {args.__dict__}')

    # Print GPU usage and set GPU visibility
    if args.num_gpus > 0:
        gpu_usage_and_visibility(args.cuda_visible_devices, args.submit)


def get_epoch_ckpt_path(exp_dir_trial: str, load_epoch: int) -> str:
    """
    Get the checkpoint path based on the epoch number.

    Argument/s:
        exp_dir_trial - Experiment directory for the trial (where the checkpoints are saved).
        load_epoch - epoch to load.

    Returns:
        Path to the epoch's checkpoint.
    """
    try:
        ckpt_path = glob.glob(os.path.join(exp_dir_trial, "*epoch=" + str(load_epoch) + "*.ckpt"))
        assert len(ckpt_path) == 1, f'Multiple checkpoints for epoch {load_epoch}: {ckpt_path}.'
    except:
        raise ValueError(
            "Epoch {} is not in the checkpoint directory.".format(str(load_epoch))
        )
    return ckpt_path[0]


def get_best_ckpt_path(exp_dir_trial: str, monitor_mode: str) -> str:
    """
    Get the best performing checkpoint from the experiment directory.

    Argument/s:
        exp_dir_trial - Experiment directory for the trial (where the checkpoints are saved).
        monitor_mode - Metric monitoring mode, either "min" or "max".

    Returns:
        Path to the epoch's checkpoint.
    """

    ckpt_list = glob.glob(os.path.join(exp_dir_trial, '*=*=*.ckpt'))

    if not ckpt_list:
        raise ValueError(f'No checkpoints exist in the checkpoint directory: {exp_dir_trial}.')

    scores = [
        re.findall(r"[-+]?\d*\.\d+|\d+", i.rsplit('=', 1)[1])[0] for i in ckpt_list
    ]

    if monitor_mode == 'max':
        ckpt_path = ckpt_list[np.argmax(scores)]
    elif monitor_mode == 'min':
        ckpt_path = ckpt_list[np.argmin(scores)]
    else:
        raise ValueError("'monitor_mode' must be max or min, not {}.".format(monitor_mode))
    return ckpt_path


def get_test_ckpt_path(
        exp_dir_trial: str, monitor_mode: str, test_epoch: Optional[int] = None, test_ckpt_path: Optional[str] = None,
) -> str:
    """
    Get the test checkpoint.

    Argument/s:
        exp_dir_trial - Experiment directory for the trial (where the checkpoints are saved).
        monitor_mode - Metric motitoring mode, either "min" or "max".
        test_epoch - epoch to test.
        test_ckpt_path - path to checkpoint to be tested.

    Returns:
        Path to the epoch's checkpoint.
    """

    set_options_bool = list(map(bool, [test_epoch is not None, test_ckpt_path]))
    assert set_options_bool.count(True) <= 1, f'Both "test_epoch" and "test_ckpt_path" cannot be set.'

    if test_ckpt_path:
        assert os.path.isfile(test_ckpt_path), f'File does not exist: {test_ckpt_path}.'
        ckpt_path = test_ckpt_path
    elif test_epoch is not None:
        ckpt_path = get_epoch_ckpt_path(exp_dir_trial, load_epoch=test_epoch)
    else:
        ckpt_path = get_best_ckpt_path(exp_dir_trial, monitor_mode)

    return ckpt_path


def write_test_ckpt_path(ckpt_path: str, exp_dir_trial: str):
    """
    Write ckpt_path used for testing to a text file.

    Argument/s:
        ckpt_path - path to the checkpoint of the epoch that scored
            highest for the given validation metric.
        exp_dir_trial - Experiment directory for the trial.
    """
    with open(os.path.join(exp_dir_trial, 'test_ckpt_path.txt'), 'a') as f:
        f.write(ckpt_path + "\n")


def resume_from_ckpt_path(
        exp_dir_trial: str,
        resumable: bool = False,
        resume_epoch: int = None,
        resume_ckpt_path: str = None,
):
    """
    Resume training from the specified checkpoint.

    Argument/s:
        exp_dir_trial - Experiment directory for the trial (where the checkpoints are saved).
        resumable - if training is automatically resumable (i.e., provide last.ckpt as the path).
        resume_epoch - get the path of the checkpoint for a given epoch.
        resume_ckpt_path - outright provide the checkpoint path.

    Returns:
          ckpt_path - path to a checkpoint.
    """

    ckpt_path = None

    options = ['resumable', 'resume_epoch', 'resume_ckpt_path']
    set_options_bool = list(map(bool, [resumable, resume_epoch is not None, resume_ckpt_path]))
    set_options = [j for i, j in enumerate(options) if set_options_bool[i]]
    assert set_options_bool.count(True) in [0, 1], \
        f'A maximum of one of these options can be set: {options}. The following are set: {set_options}.'

    if resumable:
        if os.path.isfile(os.path.join(exp_dir_trial, 'last.ckpt')):
            ckpt_path = os.path.join(exp_dir_trial, 'last.ckpt')
        else:
            warnings.warn('last.ckpt does not exist, starting training from epoch 0.')
    elif resume_epoch is not None:
        ckpt_path = get_epoch_ckpt_path(exp_dir_trial, load_epoch=resume_epoch)
    elif resume_ckpt_path:
        ckpt_path = resume_ckpt_path

    if ckpt_path is not None:
        print('Resuming training from {}.'.format(ckpt_path))

    return ckpt_path


def gpu_usage_and_visibility(cuda_visible_devices: Optional[str] = None, submit: bool = False):
    """
    Prints out GPU utilisation and sets the visibility of the GPUs to the job.

    Argument/s:
        cuda_visible_devices - which GPUs are visible to the job.
        submit - if the job is being submitted to a cluster manager, ignore GPU visibility requirements.
    """

    # Print GPU usage
    for gpu in GPUtil.getGPUs():
        print(f'Initial utilisation on GPU:{gpu.id} is {gpu.memoryUtil:0.3f}.')

    # Determine the visibility of devices
    if cuda_visible_devices and (not submit):
        os.environ['CUDA_VISIBLE_DEVICES'] = cuda_visible_devices
        print(f'CUDA_VISIBLE_DEVICES: {cuda_visible_devices}')


class suppress_stdout_stderr(object):
    """
    Suppress STDOUT and STDERR.
    """

    def __init__(self):
        # Open a pair of null files
        self.null_fds = [os.open(os.devnull, os.O_RDWR) for x in range(2)]

        # Save the actual stdout (1) and stderr (2) file descriptors.
        self.save_fds = [os.dup(1), os.dup(2)]

    def __enter__(self):
        # Assign the null pointers to stdout and stderr.
        os.dup2(self.null_fds[0], 1)
        os.dup2(self.null_fds[1], 2)

    def __exit__(self, *_):
        # Re-assign the real stdout/stderr back to (1) and (2)
        os.dup2(self.save_fds[0], 1)
        os.dup2(self.save_fds[1], 2)

        # Close all file descriptors
        for fd in self.null_fds + self.save_fds:
            os.close(fd)



from argparse import Namespace
from dlhpcstarter.command_line_arguments import read_command_line_arguments
from dlhpcstarter.utils import importer, gpu_usage_and_visibility, load_config_and_update_args
from dlhpcstarter.cluster import ClusterSubmit
from typing import Callable
import sys


def main() -> None:
    """
    The main function handles the jobs for a task. It does the following in order:
        1. Get command line arguments using argparse (or input your own arguments object to main()).
        2. Import the 'stages' function for the task.
        3. Load the configuration for the job and add it to 'args'.
        4. Submit the job to the cluster manager (or run locally).
    """

    if sys.path[0] != '':
        sys.path.insert(0, '')

    """    
    1. Get command line arguments using argparse for the job:
    """
    args = read_command_line_arguments()

    """
    2. Import the 'stages' function for the task:
    
        Imports the function that handles the training and testing stages for the task. This is the stages() function
        defined in the task's stages.py. The model is also defined in the stages function based on the configuration.
    
        For example: stages() in task.cifar10.stages 
    
    """
    stages_fnc = importer(definition='stages', module='.'.join(['task', args.task, 'stages']))

    """
    3. Load the configuration for the job and add it to 'args':
    
        This contains the paths, model configuration, the training and test configuration, the device & cluster manager 
        configuration.
    """
    load_config_and_update_args(args=args, print_args=True)

    """
    4. Submit the job to the cluster manager (or run locally if args.submit = False).
    """
    submit(args=args, stages_fnc=stages_fnc)


def submit(stages_fnc: Callable, args: Namespace):
    """
    Submit the job to the cluster manager, e.g., SLURM, or run the job locally, i.e., args.submit = False.

    Argument/s:
        stages_fnc - function that handles the creation, training, and testing of a model.
        args - arguments containing the configuration for the job and the model
    """

    if not args.submit:

        # Run locally
        stages_fnc(args)

    else:

        # Cluster object
        cluster = ClusterSubmit(
            fnc_kwargs=args,
            fnc=stages_fnc,
            save_dir=args.exp_dir_trial,
            time_limit=args.time_limit,
            num_gpus=args.num_gpus,
            num_nodes=args.num_nodes,
            num_workers=args.num_workers,
            memory=args.memory,
            python_cmd='',
            entrypoint='dlhpcstarter',
        )

        # Cluster commands
        cluster.add_manager_cmd(cmd='ntasks-per-node', value=args.num_gpus if args.num_gpus else 1)

        # Source virtual environment
        cluster.add_command('source ' + args.venv_path)

        # Debug flags
        cluster.add_command('export NCCL_DEBUG=INFO')
        cluster.add_command('export PYTHONFAULTHANDLER=1')

        # Add commands to the cluster manager submission script
        if 'cluster_manager_script_commands' in args:
            for i in args.cluster_manager_script_commands:
                cluster.add_command(i)

        # Add cluster manager commands
        if 'cluster_manager_commands' in args:
            for i in args.cluster_manager_commands:
                cluster.add_manager_cmd(cmd=i[0], value=i[1])

        # Request the quality of service for the job
        if args.qos:
            cluster.add_manager_cmd(cmd='qos', value=args.qos)

        # Email job status
        cluster.notify_job_status(email=args.email, on_done=True, on_fail=True)

        # Submit job to workload manager
        job_display_name = args.task + '_' + args.config

        if args.trial is not None:
            job_display_name = job_display_name + f'_trial_{args.trial}'

        cluster.submit(job_display_name=job_display_name + '_')


if __name__ == '__main__':
    main()



from argparse import Namespace
from subprocess import call
from typing import Callable, Optional
import datetime
import os
import signal
import sys
import traceback


class ClusterSubmit(object):
    """
    Submits a job to a cluster manager.
    """
    def __init__(
            self,
            fnc: Callable,
            fnc_kwargs: Namespace,
            save_dir: str,
            log_err: bool = True,
            log_out: bool = True,
            manager: str = 'slurm',
            time_limit: str = '02:00:00',
            begin: str = 'now',
            num_gpus: int = 0,
            num_nodes: int = 1,
            num_workers: int = 1,
            memory: str = '16GB',
            no_srun: bool = False,
            python_cmd: str = 'python3',
            entrypoint: Optional[str] = None,
            resubmit: bool = True,
    ):
        """
        Argument/s:
            fnc - function for the job.
            fnc_kwargs - keyword arguments for the function.
            save_dir -  directory where the cluster manager script, stdout, and stderr are saved.
            log_err - log standard error.
            log_out - log standard output.
            manager - 'slurm' (needs to be modified for other cluster managers, e.g., PBS).
            time_limit - time limit for job.
            begin - when to begin the job.
            num_gpus - number of gpus.
            num_nodes - number of nodes.
            num_workers - number of workers per GPU.
            memory - minimum memory amount.
            no_srun - don't use 'srun'.
            python_cmd - python command name.
            entrypoint - entrypoint to use instead of a script.
            resubmit - automatically resubmit job just before timout.
        """
        self.fnc = fnc
        self.fnc_kwargs = fnc_kwargs
        self.exp_dir = save_dir
        self.log_err = log_err
        self.log_out = log_out
        self.manager = manager
        self.time_limit = time_limit
        self.begin = begin
        self.num_gpus = num_gpus
        self.num_nodes = num_nodes
        self.num_workers = num_workers
        self.memory = memory
        self.no_srun = no_srun
        self.python_cmd = python_cmd
        self.entrypoint = entrypoint
        self.resubmit = resubmit

        self.script_name = os.path.realpath(sys.argv[0])
        self.manager_commands = []
        self.commands = []

        self.run_cmd = {
            'slurm': 'sbatch',
        }

        self.is_from_manager_object = bool(vars(fnc_kwargs)["slurm_cmd_path"])

    def add_manager_cmd(self, cmd=None, value=None):
        self.manager_commands.append((cmd, value))

    def add_command(self, cmd):
        self.commands.append(cmd)

    def notify_job_status(self, email, on_done, on_fail):
        self.email = email
        self.notify_on_end = on_done
        self.notify_on_fail = on_fail

    def submit(self, job_display_name=None):
        self.job_display_name = job_display_name
        self.log_dir()
        if self.is_from_manager_object:
            self.run()
        else:
            scripts_path = os.path.join(self.exp_dir, 'manager_scripts')
            self.schedule_experiment(self.get_max_session_number(scripts_path))

    def run(self):
        if self.resubmit:
            print('Setting signal to automatically requeue the job before timeout.')
            signal.signal(signal.SIGUSR1, self.sig_handler)
            signal.signal(signal.SIGTERM, self.term_handler)
        else:
            print("Automatic requeuing has not been set. The job will not be requeued after timeout.")
        try:
            self.fnc(self.fnc_kwargs)

        except Exception as e:
            print('Caught exception in worker thread', e)
            traceback.print_exc()
            raise SystemExit

    def schedule_experiment(self, session):
        timestamp = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        timestamp = 'session_{}_{}'.format(session, timestamp)

        # Generate and save cluster manager script
        manager_cmd_script_path = os.path.join(self.manager_files_log_path, '{}.sh'.format(timestamp))
        if self.manager == 'slurm':
            manager_cmd = self.build_slurm_command(manager_cmd_script_path, timestamp, session)
        else:
            raise ValueError(f'{self.manager} is not a valid manager.')
        self.save_manager_cmd(manager_cmd, manager_cmd_script_path)

        # Run script to launch job
        print('\nLaunching experiment...')
        result = call('{} {}'.format(self.run_cmd[self.manager], manager_cmd_script_path), shell=True)
        if result == 0:
            print(f'Launched experiment {manager_cmd_script_path}.')
        else:
            print('Launch failed...')

    def call_resume(self):
        job_id = os.environ['SLURM_JOB_ID']
        cmd = 'scontrol requeue {}'.format(job_id)
        print(f'\nRequeing job {job_id}...')
        result = call(cmd, shell=True)
        if result == 0:
            print(f'Requeued job {job_id}.')
        else:
            print('Requeue failed...')
        os._exit(0)

    def sig_handler(self, signum, frame):
        print(f"Caught signal: {signum}")
        self.call_resume()

    def term_handler(self, signum, frame):
        print("Bypassing sigterm.")

    def save_manager_cmd(self, manager_cmd, manager_cmd_script_path):
        with open(manager_cmd_script_path, mode='w') as file:
            file.write(manager_cmd)

    def get_max_session_number(self, path):
        files = os.listdir(path)
        session_files = [f for f in files if 'session_' in f]
        if len(session_files) > 0:
            sessions = [int(f_name.split('_')[1]) for f_name in session_files]
            max_session = max(sessions)
            return max_session + 1
        else:
            return 0

    def log_dir(self):
        out_path = os.path.join(self.exp_dir)
        self.exp_dir = out_path
        if not os.path.exists(out_path):
            os.makedirs(out_path)
        self.manager_files_log_path = os.path.join(out_path, 'manager_scripts')
        if not os.path.exists(self.manager_files_log_path):
            os.makedirs(self.manager_files_log_path)
        if self.log_err:
            err_path = os.path.join(out_path, 'error_logs')
            if not os.path.exists(err_path):
                os.makedirs(err_path)
            self.err_log_path = err_path
        if self.log_out:
            out_path = os.path.join(out_path, 'out_logs')
            if not os.path.exists(out_path):
                os.makedirs(out_path)
            self.out_log_path = out_path

    def build_slurm_command(self, manager_cmd_script_path, timestamp, session):

        sub_commands = ['#!/bin/bash -l']
        sub_commands.append('#SBATCH --job-name={}'.format('{}session_{}'.format(self.job_display_name, session)))

        if self.log_out:
            out_path = os.path.join(self.out_log_path, '{}_%j.out'.format(timestamp))
            sub_commands.append('#SBATCH --output={}'.format(out_path))

        if self.log_err:
            err_path = os.path.join(self.err_log_path, '{}_%j.err'.format(timestamp))
            sub_commands.append('#SBATCH --error={}'.format(err_path))
        sub_commands.append(f'#SBATCH --time={self.time_limit:s}')

        if self.begin != "now":
            sub_commands.append('#SBATCH --begin={}'.format(self.begin))

        if self.num_gpus:
            sub_commands.append('#SBATCH --gres=gpu:{}'.format(self.num_gpus))

        if self.num_workers > 0:
            sub_commands.append('#SBATCH --cpus-per-task={}'.format(self.num_workers))

        sub_commands.append('#SBATCH --nodes={}'.format(self.num_nodes))
        sub_commands.append('#SBATCH --mem={}'.format(self.memory))
        sub_commands.append(f'#SBATCH --signal=USR1@{6 * 60}')

        mail_type = []
        if self.notify_on_end:
            mail_type.append('END')
        if self.notify_on_fail:
            mail_type.append('FAIL')
        if len(mail_type) > 0 and self.email is not None:
            sub_commands.append('#SBATCH --mail-type={}'.format(','.join(mail_type)))
            sub_commands.append('#SBATCH --mail-user={}'.format(self.email))

        for (cmd, value) in self.manager_commands:
            if value:
                sub_commands.append('#SBATCH --{}={}'.format(cmd, value))
            else:
                sub_commands.append('#SBATCH --{}'.format(cmd))

        sub_commands = [x.lstrip() for x in sub_commands]

        for cmd in self.commands:
            sub_commands.append(cmd)

        args_string = self.args_to_string(self.fnc_kwargs)
        args_string = '{} --{} {}'.format(args_string, "slurm_cmd_path", manager_cmd_script_path)

        if self.entrypoint:
            cmd = f'{self.entrypoint} {args_string}'
        else:
            cmd = '{} {} {}'.format(self.python_cmd, self.script_name, args_string)
        if not self.no_srun:
            cmd = 'srun ' + cmd
        sub_commands.append(cmd)

        return '\n'.join(sub_commands)

    def args_to_string(self, args):
        params = []
        for k, v in vars(args).items():
            if v is not None:
                if self.escape(v):
                    cmd = '--{} \"{}\"'.format(k, v)
                else:
                    cmd = '--{} {}'.format(k, v)
                params.append(cmd)
        return ' '.join(params)

    def escape(self, v):
        v = str(v)
        return '[' in v or ';' in v or ' ' in v



# Copyright The PyTorch Lightning team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Diagnose your system and show basic information

This server mainly to get detail info for better bug reporting.

"""

import os
import platform
import re
import sys

import numpy
import torch
import tqdm

import pytorch_lightning  # noqa: E402

LEVEL_OFFSET = "\t"
KEY_PADDING = 20


def run_and_parse_first_match(run_lambda, command, regex):
    """Runs command using run_lambda, returns the first regex match if it exists"""
    rc, out, _ = run_lambda(command)
    if rc != 0:
        return None
    match = re.search(regex, out)
    if match is None:
        return None
    return match.group(1)


def get_running_cuda_version(run_lambda):
    return run_and_parse_first_match(run_lambda, "nvcc --version", r"V(.*)$")


def info_system():
    return {
        "OS": platform.system(),
        "architecture": platform.architecture(),
        "version": platform.version(),
        "processor": platform.processor(),
        "python": platform.python_version(),
    }


def info_cuda():
    return {
        "GPU": [
            torch.cuda.get_device_name(i) for i in range(torch.cuda.device_count())
        ],
        # 'nvidia_driver': get_nvidia_driver_version(run_lambda),
        "available": torch.cuda.is_available(),
        "version": torch.version.cuda,
    }


def info_packages():
    return {
        "numpy": numpy.__version__,
        "pyTorch_version": torch.__version__,
        "pyTorch_debug": torch.version.debug,
        "pytorch-lightning": pytorch_lightning.__version__,
        "tqdm": tqdm.__version__,
    }


def nice_print(details, level=0):
    lines = []
    for k in sorted(details):
        key = f"* {k}:" if level == 0 else f"- {k}:"
        if isinstance(details[k], dict):
            lines += [level * LEVEL_OFFSET + key]
            lines += nice_print(details[k], level + 1)
        elif isinstance(details[k], (set, list, tuple)):
            lines += [level * LEVEL_OFFSET + key]
            lines += [(level + 1) * LEVEL_OFFSET + "- " + v for v in details[k]]
        else:
            template = "{:%is} {}" % KEY_PADDING
            key_val = template.format(key, details[k])
            lines += [(level * LEVEL_OFFSET) + key_val]
    return lines


def main():
    details = {
        "System": info_system(),
        "CUDA": info_cuda(),
        "Packages": info_packages(),
    }
    lines = nice_print(details)
    text = os.linesep.join(lines)
    print(text)


if __name__ == "__main__":
    main()
